use std::fs::File;
use std::io::Read;
use std::env;

use rand::Rng;

struct RndIter {
    to_read: usize,
    chunk_size: usize,
    urandom: File,
    buff: Vec<u8>,
    cur_read: usize,
    must_read: bool,
}

impl RndIter {
    fn new(count: usize, chunk_size: usize) -> RndIter {
        RndIter {   to_read: count,
                    chunk_size: chunk_size,
                    urandom: File::open("/dev/urandom")
                        .expect("Unable to open /dev/urandom"),
                    buff: vec![0;chunk_size],
                    cur_read: 0,
                    must_read: true,
                }
    }
}

impl Iterator for RndIter {

    type Item = u8;

    fn next(&mut self) -> Option<Self::Item> {

        if self.must_read {
            if self.to_read == 0 {
                return None;
            }
            if self.to_read < self.chunk_size {
                self.buff = vec![0;self.to_read];
                self.to_read = 0;
            } else {
                self.to_read -=  self.chunk_size;
            }
            self.urandom.read(&mut self.buff)
                .expect("Unable to read /dev/urandom");
            self.cur_read = 0;
            self.must_read = false;
        }

        if self.cur_read == self.buff.len() - 1 {
            self.must_read = true;
        }
        let ret = self.buff[self.cur_read];
        self.cur_read += 1;
        return Some(ret);
    }
}

fn std_rand(resolution: u8, iterations: usize, result: &mut Vec<u32>) {
    let mut rnd_thread = rand::thread_rng();

    let mut rnd: u8;
    for _  in 0..iterations {
        rnd = rnd_thread.gen_range(0..resolution);
        result[rnd as usize]+=1;
    }
}

fn mod_rand(resolution: u8, iterations: usize, result: &mut Vec<u32>) {
    let rnd_gen = RndIter::new(iterations, 4096);
    let mut idx: usize;

    for value in rnd_gen {
        idx = (value % resolution).into();
        result[idx] += 1;
    }
}

fn prop_rand(resolution: u8, iterations: usize, result: &mut Vec<u32>) {
    let rnd_gen = RndIter::new(iterations, 4096);
    let mut idx: usize;

    for value in rnd_gen {
        idx = ((value as u16 * resolution as u16) / 256).into();
        result[idx] += 1;
    }
}

fn print_results(results: &[Vec<u32>;3], resolution: u8, iterations: usize, relative: bool) {
    for i in 0..resolution {
        let idx: usize = i.into();
        if relative {
            println!("{} {} {} {} {}", resolution, i as f64/resolution as f64,
                (results[0][idx] as f64 * resolution as f64) / iterations as f64,
                (results[1][idx] as f64 * resolution as f64) / iterations as f64,
                (results[2][idx] as f64 * resolution as f64) / iterations as f64);
        } else {
            println!("{} {} {} {} {}", resolution, i,
                results[0][idx], results[1][idx], results[2][idx]);
        }
    }
}

fn usage(progname: String) {
    println!("Usage {} [-i ITERATIONS] [-r RESOLUTION] [-R]", progname);
    println!("\t-i, --iterations Iteration count");
    println!("\t-r, --resolution Dice faces count");
    println!("\t-R, --relative   Print relative count instead of absolute");
}

fn main() {
    
    let mut iterations: usize = 0x10000;
    let mut resolution: u8 = 100;
    let mut relative:  bool = false;

    let mut args = env::args();
    let _argv0 = args.next().expect("Fatal error");

    loop {
        match args.next() {
            Some(arg) =>  {
                if arg == "-h" || arg == "--help"{
                    usage(_argv0);
                    return;
                } else if arg == "-r" || arg == "--resolution" {
                    match args.next() {
                        Some(x) => resolution = x.parse().expect("Resolution in not a number"),
                        None => {
                            println!("Missing argument to -r");
                            return;
                        }
                    }
                } else if arg == "-i" || arg == "--iterations" {
                    match args.next() {
                        Some(x) => iterations = x.parse().expect("Iteration is not a number"),
                        None => {
                            println!("Missing argument to -i");
                            return;
                        }
                    }
                } else if arg == "-R" || arg  == "--relative" {
                    relative = true;
                } else {
                    usage(_argv0);
                    panic!("Invalid argument '{}'", arg);
                }
            },
            None => break,
        }
    }

    if relative {
        iterations *= resolution as usize;
    }

    let mut results: [Vec<u32>;3] = [
        vec![0;resolution.into()],
        vec![0;resolution.into()],
        vec![0;resolution.into()],
    ];

    std_rand(resolution, iterations, &mut results[0]);
    mod_rand(resolution, iterations, &mut results[1]);
    prop_rand(resolution, iterations, &mut results[2]);

    print_results(&results, resolution, iterations, relative);
}

